"""
Tool to extract and pretty-print PDF annotations for reviewing.
"""

__version__ = '0.5'

import bisect
import collections
import itertools
import logging
import typing as typ

from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter
from pdfminer.pdfpage import PDFPage
from pdfminer.layout import (LAParams, LTAnno, LTChar, LTComponent, LTContainer, LTFigure, LTItem,
                             LTPage, LTTextBox, LTTextLine)
from pdfminer.converter import PDFLayoutAnalyzer
from pdfminer.pdfparser import PDFParser
from pdfminer.pdfdocument import PDFDocument, PDFNoOutlines
from pdfminer.psparser import PSLiteralTable, PSLiteral
from pdfminer import pdftypes
import pdfminer.settings
import pdfminer.utils

from .types import Page, Outline, AnnotationType, Annotation, Document, RGB
from .utils import cleanup_text, decode_datetime

pdfminer.settings.STRICT = False

logger = logging.getLogger('pdfannots')

ANNOT_SUBTYPES: typ.Dict[PSLiteral, AnnotationType] = {
    PSLiteralTable.intern(e.name): e for e in AnnotationType}
"""Mapping from PSliteral to our own enumerant, for supported annotation types."""

IGNORED_ANNOT_SUBTYPES = \
    frozenset(PSLiteralTable.intern(n) for n in (
        'Link',   # Links are used for internal document links (e.g. to other pages).
        'Popup',  # Controls the on-screen appearance of other annotations. TODO: we may want to
                  # check for an optional 'Contents' field for alternative human-readable contents.
    ))
"""Annotation types that we ignore without issuing a warning."""


def _mkannotation(
    pa: typ.Dict[str, typ.Any],
    page: Page
) -> typ.Optional[Annotation]:
    """
    Given a PDF annotation, capture relevant fields and construct an Annotation object.

    Refer to Section 8.4 of the PDF reference (version 1.7).
    """

    subtype = pa.get('Subtype')
    annot_type = None
    assert isinstance(subtype, PSLiteral)
    try:
        annot_type = ANNOT_SUBTYPES[subtype]
    except KeyError:
        pass

    if annot_type is None:
        if subtype not in IGNORED_ANNOT_SUBTYPES:
            logger.warning("Unsupported %s annotation ignored on %s", subtype.name, page)
        return None

    contents = pa.get('Contents')
    if contents is not None:
        # decode as string, normalise line endings, replace special characters
        contents = cleanup_text(pdfminer.utils.decode_text(contents))

    rgb: typ.Optional[RGB] = None
    color = pdftypes.resolve1(pa.get('C'))
    if color:
        if (isinstance(color, list)
                and len(color) == 3
                and all(isinstance(e, (int, float)) and 0 <= e <= 1 for e in color)):
            rgb = RGB(*color)
        else:
            logger.warning("Invalid color %s in annotation on %s", color, page)

    # Rect defines the location of the annotation on the page
    rect = pdftypes.resolve1(pa.get('Rect'))

    # QuadPoints are defined only for "markup" annotations (Highlight, Underline, StrikeOut,
    # Squiggly, Caret), where they specify the quadrilaterals (boxes) covered by the annotation.
    quadpoints = pdftypes.resolve1(pa.get('QuadPoints'))

    author = pdftypes.resolve1(pa.get('T'))
    if author is not None:
        author = pdfminer.utils.decode_text(author)

    name = pdftypes.resolve1(pa.get('NM'))
    if name is not None:
        name = pdfminer.utils.decode_text(name)

    created = None
    dobj = pa.get('CreationDate')
    # some pdf apps set modification date, but not creation date
    dobj = dobj or pa.get('ModDate')
    # poppler-based apps (e.g. Okular) use 'M' for some reason
    dobj = dobj or pa.get('M')
    createds = pdftypes.resolve1(dobj)
    if createds is not None:
        createds = pdfminer.utils.decode_text(createds)
        created = decode_datetime(createds)

    in_reply_to = pa.get('IRT')
    is_group = False
    if in_reply_to is not None:
        reply_type = pa.get('RT')
        if reply_type is PSLiteralTable.intern('Group'):
            is_group = True
        elif not (reply_type is None or reply_type is PSLiteralTable.intern('R')):
            logger.warning("Unexpected RT=%s, treated as R", reply_type)

    return Annotation(page, annot_type, quadpoints=quadpoints, rect=rect, name=name,
                      contents=contents, author=author, created=created, color=rgb,
                      in_reply_to_ref=in_reply_to, is_group_child=is_group)


def _get_outlines(doc: PDFDocument) -> typ.Iterator[Outline]:
    """Retrieve a list of (unresolved) Outline objects for all recognised outlines in the PDF."""

    def _resolve_dest(dest: typ.Any) -> typ.Any:
        if isinstance(dest, pdftypes.PDFObjRef):
            dest = pdftypes.resolve1(dest)
        if isinstance(dest, bytes):
            dest = pdftypes.resolve1(doc.get_dest(dest))
        elif isinstance(dest, PSLiteral):
            dest = pdftypes.resolve1(doc.get_dest(dest.name))
        if isinstance(dest, dict):
            dest = dest['D']
        return dest

    for (_, title, destname, actionref, _) in doc.get_outlines():
        if destname is None and actionref:
            action = pdftypes.resolve1(actionref)
            if isinstance(action, dict):
                subtype = action.get('S')
                if subtype is PSLiteralTable.intern('GoTo'):
                    destname = action.get('D')
        if destname is None:
            continue
        dest = _resolve_dest(destname)

        # consider targets of the form [page /XYZ left top zoom]
        if dest[1] is PSLiteralTable.intern('XYZ'):
            (pageref, _, targetx, targety) = dest[:4]

            if not isinstance(pageref, (int, pdftypes.PDFObjRef)):
                logger.warning("Unsupported pageref in outline: %s", pageref)
            else:
                if targetx is None or targety is None:
                    # Treat as a general reference to the page
                    target = None
                else:
                    target = (targetx, targety)
                    if not all(isinstance(v, (int, float)) for v in target):
                        logger.warning("Unsupported target in outline: (%r, %r)", targetx, targety)
                        target = None

                yield Outline(title, pageref, target)


class _PDFProcessor(PDFLayoutAnalyzer):
    """
    PDF processor class.

    This class encapsulates our primary interface with pdfminer's page layout logic. It is used
    to define a logical order for the objects we care about (Annotations and Outlines) on a page,
    and to capture the text that annotations may refer to.
    """

    CONTEXT_CHARS = 256
    """Maximum number of recent characters to keep as context."""

    page: typ.Optional[Page]                # Page being processed.
    charseq: int                            # Character sequence number within the page.
    compseq: int                            # Component sequence number within the page.
    recent_text: typ.Deque[str]             # Rotating buffer of recent text, for context.
    _lasthit: typ.FrozenSet[Annotation]     # Annotations hit by the most recent character.
    _curline: typ.Set[Annotation]           # Annotations hit somewhere on the current line.

    # Stores annotations that are subscribed to receive their post-annotation
    # context. The first element of each tuple, on which the list is sorted, is
    # the sequence number of the last character to hit the annotation.
    context_subscribers: typ.List[typ.Tuple[int, Annotation]]

    def __init__(self, rsrcmgr: PDFResourceManager, laparams: LAParams):
        super().__init__(rsrcmgr, laparams=laparams)
        self.page = None
        self.recent_text = collections.deque(maxlen=self.CONTEXT_CHARS)
        self.context_subscribers = []
        self.clear()

    def clear(self) -> None:
        """Reset our internal per-page state."""
        self.charseq = 0
        self.compseq = 0
        self.recent_text.clear()
        self.context_subscribers.clear()
        self._lasthit = frozenset()
        self._curline = set()

    def set_page(self, page: Page) -> None:
        """Prepare to process a new page. Must be called prior to processing."""
        assert self.page is None
        self.page = page

    def receive_layout(self, ltpage: LTPage) -> None:
        """Callback from PDFLayoutAnalyzer superclass. Called once with each laid-out page."""
        assert self.page is not None

        # Re-initialise our per-page state
        self.clear()

        # Render all the items on the page
        self.render(ltpage)

        # If we still have annotations needing context, give them whatever we have
        for (charseq, annot) in self.context_subscribers:
            available = self.charseq - charseq
            annot.post_context = ''.join(self.recent_text[n] for n in range(-available, 0))

        self.page = None

    def update_pageseq(self, component: LTComponent) -> bool:
        """Assign sequence numbers for objects on the page based on the nearest line of text.
        Returns True if we need to recurse on smaller sub-components (e.g. characters)."""
        assert self.page is not None
        self.compseq += 1

        hits = 0
        for x in itertools.chain(self.page.annots, self.page.outlines):
            if x.update_pageseq(component, self.compseq):
                hits += 1

        # If we have assigned the same sequence number to multiple objects, and there exist smaller
        # sub-components (e.g. characters within a line), we'll recurse on those assigning sequence
        # numbers to sub-components to disambiguate the hits, but first we must forget about the
        # current sequence number.
        # NB: This could be done more efficiently -- we really only need to disambiguate conflicts
        # that still exist after processing *all* the line-level components on the same page, but
        # that would require multiple rendering passes.
        if hits > 1 and isinstance(component, LTContainer) and len(component) > 1:
            for x in itertools.chain(self.page.annots, self.page.outlines):
                x.discard_pageseq(self.compseq)
            return True

        return False

    def test_boxes(self, item: LTComponent) -> None:
        """Update the set of annotations whose boxes intersect with the area of the given item."""
        assert self.page is not None
        hits = frozenset(a for a in self.page.annots if a.boxes
                         and any(b.hit_item(item) for b in a.boxes))
        self._lasthit = hits
        self._curline.update(hits)

    def capture_context(self, text: str) -> None:
        """Store the character for use as context, and update subscribers if required."""
        self.recent_text.append(text)
        self.charseq += 1

        # Notify subscribers for whom this character provides the full post-context.
        while self.context_subscribers:
            (charseq, annot) = self.context_subscribers[0]
            assert charseq < self.charseq
            if charseq == self.charseq - self.CONTEXT_CHARS:
                annot.set_post_context(''.join(self.recent_text))
                self.context_subscribers.pop(0)
            else:
                assert charseq > self.charseq - self.CONTEXT_CHARS
                break

    def capture_char(self, text: str) -> None:
        """Capture a character."""
        self.capture_context(text)

        if text == '\n':
            # "Broadcast" newlines to _all_ annotations that received any text on the
            # current line, in case they see more text on the next line, even if the
            # most recent character on the line was not covered by their boxes.
            for a in self._curline:
                a.capture('\n')
            self._curline = set()
        else:
            # Broadcast the character to annotations that include it.
            for a in self._lasthit:
                last_charseq = a.last_charseq
                a.capture(text, self.charseq)

                if a.wants_context():
                    if a.has_context():
                        # We already gave the annotation the pre-context, so it is subscribed.
                        # Locate and remove the annotation's existing context subscription.
                        assert last_charseq != 0
                        i = bisect.bisect_left(self.context_subscribers, (last_charseq,))
                        assert 0 <= i < len(self.context_subscribers)
                        while True:
                            (found_charseq, found_annot) = self.context_subscribers[i]
                            assert found_charseq == last_charseq
                            if found_annot is a:
                                self.context_subscribers.pop(i)
                                break
                            i += 1
                            assert i < len(self.context_subscribers)

                    else:
                        # This is the first hit for the annotation, so set the pre-context.
                        assert last_charseq == 0
                        assert len(a.text) != 0
                        pre_context = ''.join(
                            self.recent_text[n] for n in range(len(self.recent_text) - 1))
                        a.set_pre_context(pre_context)

                    # Subscribe this annotation for post-context.
                    self.context_subscribers.append((self.charseq, a))

    def render(self, item: LTItem, pageseq_nested: bool = False) -> None:
        """
        Helper for receive_layout, called recursively for every item on a page, in layout order.

        Ref: https://pdfminersix.readthedocs.io/en/latest/topic/converting_pdf_to_text.html
        """
        # Assign sequence numbers to items on the page based on their proximity to lines of text or
        # to figures (which may contain bare LTChar elements).
        if isinstance(item, (LTTextLine, LTFigure)) or (
                pageseq_nested and isinstance(item, LTComponent)):
            pageseq_nested = self.update_pageseq(item)

        # If it's a container, recurse on nested items.
        if isinstance(item, LTContainer):
            for child in item:
                self.render(child, pageseq_nested)

            # After the children of a text box, capture the end of the final
            # line (logic derived from pdfminer.converter.TextConverter).
            if isinstance(item, LTTextBox):
                self.capture_char('\n')

        # Each character is represented by one LTChar, and we must handle
        # individual characters (not higher-level objects like LTTextLine)
        # so that we can capture only those covered by the annotation boxes.
        elif isinstance(item, LTChar):
            self.test_boxes(item)
            self.capture_char(item.get_text())

        # LTAnno objects capture whitespace not explicitly encoded in
        # the text. They don't have an (X,Y) position -- we treat them
        # the same as the most recent character.
        elif isinstance(item, LTAnno):
            self.capture_char(item.get_text())


def process_file(
    file: typ.BinaryIO,
    *,  # Subsequent arguments are keyword-only
    columns_per_page: typ.Optional[int] = None,
    emit_progress_to: typ.Optional[typ.TextIO] = None,
    laparams: LAParams = LAParams()
) -> Document:
    """
    Process a PDF file, extracting its annotations and outlines.

    Arguments:
        file                Handle to PDF file
        columns_per_page    If set, overrides PDF Miner's layout detect with a fixed page layout
        emit_progress_to    If set, file handle (e.g. sys.stderr) to which progress is reported
        laparams            PDF Miner layout parameters
    """

    # Initialise PDFMiner state
    rsrcmgr = PDFResourceManager()
    device = _PDFProcessor(rsrcmgr, laparams)
    interpreter = PDFPageInterpreter(rsrcmgr, device)
    parser = PDFParser(file)
    doc = PDFDocument(parser)

    def emit_progress(msg: str) -> None:
        if emit_progress_to is not None:
            emit_progress_to.write(msg)
            emit_progress_to.flush()

    emit_progress(file.name)

    # Retrieve outlines if present. Each outline refers to a page, using
    # *either* a PDF object ID or an integer page number. These references will
    # be resolved below while rendering pages -- for now we insert them into one
    # of two dicts for later.
    outlines_by_pageno: typ.Dict[object, typ.List[Outline]] = collections.defaultdict(list)
    outlines_by_objid: typ.Dict[object, typ.List[Outline]] = collections.defaultdict(list)

    try:
        for o in _get_outlines(doc):
            if isinstance(o.pageref, pdftypes.PDFObjRef):
                outlines_by_objid[o.pageref.objid].append(o)
            else:
                outlines_by_pageno[o.pageref].append(o)
    except PDFNoOutlines:
        logger.info("Document doesn't include outlines (\"bookmarks\")")
    except Exception as ex:
        logger.warning("Failed to retrieve outlines: %s", ex)

    # Iterate over all the pages, constructing page objects.
    result = Document()
    for (pageno, pdfpage) in enumerate(PDFPage.create_pages(doc)):
        emit_progress(" %d" % (pageno + 1))

        page = Page(pageno, pdfpage.pageid, pdfpage.label, pdfpage.mediabox, columns_per_page)
        result.pages.append(page)

        # Resolve any outlines referring to this page, and link them to the page.
        # Note that outlines may refer to the page number or ID.
        for o in (outlines_by_objid.pop(page.objid, [])
                  + outlines_by_pageno.pop(pageno, [])):
            o.resolve(page)
            page.outlines.append(o)

        # Dict from object ID (in the ObjRef) to Annotation object
        # This is used while post-processing to resolve inter-annotation references
        annots_by_objid: typ.Dict[int, Annotation] = {}

        # Construct Annotation objects, and append them to the page.
        for pa in pdftypes.resolve1(pdfpage.annots) if pdfpage.annots else []:
            if isinstance(pa, pdftypes.PDFObjRef):
                annot_dict = pdftypes.dict_value(pa)
                if annot_dict:  # Would be empty if pa is a broken ref
                    annot = _mkannotation(annot_dict, page)
                    if annot is not None:
                        page.annots.append(annot)
                        assert pa.objid not in annots_by_objid
                        annots_by_objid[pa.objid] = annot
            else:
                logger.warning("Unknown annotation: %s", pa)

        # If the page has neither outlines nor annotations, skip further processing.
        if not (page.annots or page.outlines):
            continue

        # Render the page. This captures the selected text for any annotations
        # on the page, and updates annotations and outlines with a logical
        # sequence number based on the order of text lines on the page.
        device.set_page(page)
        interpreter.process_page(pdfpage)

        # Now we have their logical order, sort the annotations and outlines.
        page.annots.sort()
        page.outlines.sort()

        # Give the annotations a chance to update their internals
        for a in page.annots:
            a.postprocess(annots_by_objid)

    emit_progress("\n")

    device.close()

    # all outlines should be resolved by now
    assert {} == outlines_by_pageno
    assert {} == outlines_by_objid

    return result



import json
import typing as typ

from . import Printer
from ..types import Annotation, Document


def annot_to_dict(
    doc: Document,
    annot: Annotation,
    remove_hyphens: bool
) -> typ.Dict[str, typ.Any]:
    """Convert an annotation to a dictionary representation suitable for JSON encoding."""
    assert annot.pos

    result = {
        "name": annot.name,
        "type": annot.subtype.name,
        "page": annot.pos.page.pageno + 1,
        "page_label": annot.pos.page.label,
        "start_xy": (annot.pos.x, annot.pos.y),
        "prior_outline": getattr(doc.nearest_outline(annot.pos), 'title', None),
        "text": annot.gettext(remove_hyphens),
        "contents": annot.contents,
        "author": annot.author,
        "created": annot.created.strftime('%Y-%m-%dT%H:%M:%S') if annot.created else None,
        "color": ('#' + annot.color.ashex()) if annot.color else None,
        "in_reply_to": annot.in_reply_to.name if annot.in_reply_to else None,
    }

    # Remove keys with None values in nested dictionary and return
    return {k: v for k, v in result.items() if v is not None}


class JsonPrinter(Printer):
    def __init__(
            self,
            *,
            remove_hyphens: bool,  # Whether to remove hyphens across a line break
            output_codec: str      # Text codec in use for output
    ) -> None:
        self.remove_hyphens = remove_hyphens
        self.seen_first = False

        # JSON must be represented as UTF-8, UTF-16, or UTF-32. If the output codec is
        # one of these, we can disable ASCII string escaping in the JSON encoder.
        self.ensure_ascii = output_codec not in ['utf-8', 'utf-16', 'utf-32']

    def end(self) -> str:
        return '\n'

    def print_file(
        self,
        filename: str,
        document: Document
    ) -> typ.Iterator[str]:
        if self.seen_first:
            # The flat array format is incompatible with multiple input files
            # TODO: Ideally we'd catch this at invocation time
            raise RuntimeError("The JSON output format does not support multiple files.")
        else:
            self.seen_first = True

        annots = [annot_to_dict(document, a, self.remove_hyphens)
                  for a in document.iter_annots(include_replies=True)]
        yield from json.JSONEncoder(indent=2, ensure_ascii=self.ensure_ascii).iterencode(annots)



from collections import defaultdict
import logging
import textwrap
import typing as typ

from . import Printer
from ..types import RGB, AnnotationType, Pos, Annotation, Document

logger = logging.getLogger('pdfannots')

MAX_CONTEXT_WORDS = 10
"""Maximum number of words returned by trim_context."""

FALLBACK_CONTEXT_WORDS = 4
"""Number of words returned by trim_context in fallback mode."""

CONTEXT_BOUNDARIES = [
    # (separator, keep_on_left, keep_on_right)
    ('. ', False, True),  # sentence boundary
    ('! ', False, True),
    ('? ', False, True),
    (': ', False, False),
    ('; ', False, False),
    ('" ', False, True),   # end of quote
    (' "', True, False),   # start of quote
    (') ', False, True),   # end of parenthesis
    (' (', True, False),   # start of parenthesis
    ('—', False, False),   # em dash
]
"""Rough approximation of natural boundaries in writing, used when searching for context."""


def trim_context(context: str, keep_right: bool) -> str:
    """
    Trim context for presentation.

    Given a potentially-long string of context preceding or following an annotation, identify
    a natural boundary at which to trim it, and return the trimmed string.

    Arguments:
        context     String of captured context
        keep_right  Whether to retain text on the right (True) or left (False) end of the string
    """
    best = None

    for (sep, keep_sep_left, keep_sep_right) in CONTEXT_BOUNDARIES:
        # search for the separator
        i = context.rfind(sep) if keep_right else context.find(sep)
        if i < 0:
            continue

        # include the separator if desired
        if (keep_right and not keep_sep_left) or (not keep_right and keep_sep_right):
            i += len(sep)

        # extract the candidate string
        candidate = context[i:] if keep_right else context[:i]

        if best is None or len(candidate) < len(best):
            best = candidate
            if len(candidate.split()) <= 1:
                break

    if best is not None and len(best.split()) <= MAX_CONTEXT_WORDS:
        return best

    # Give up and take a few words, whatever they are.
    if keep_right:
        fallback = '...' + ' '.join(context.split()[-FALLBACK_CONTEXT_WORDS:])
        if context[-1].isspace():
            fallback += context[-1]
    else:
        fallback = ' '.join(context.split()[:FALLBACK_CONTEXT_WORDS]) + '...'
        if context[0].isspace():
            fallback = context[0] + fallback

    return fallback


class MarkdownPrinter(Printer):
    BULLET_INDENT1 = " * "
    BULLET_INDENT2 = "   "
    QUOTE_INDENT = BULLET_INDENT2 + "> "

    def __init__(
        self,
        *,
        condense: bool = True,                  # Permit use of the condensed format
        page_number_offset: int = 0,            # Page number offset
        print_filename: bool = False,           # Whether to print file names
        remove_hyphens: bool = True,            # Whether to remove hyphens across a line break
        use_page_labels: bool = True,           # Whether to use page labels
        wrap_column: typ.Optional[int] = None,  # Column at which output is word-wrapped
        **kwargs: typ.Any                       # Other args, ignored
    ) -> None:
        self.page_number_offset = page_number_offset
        self.print_filename = print_filename
        self.remove_hyphens = remove_hyphens
        self.use_page_labels = use_page_labels
        self.wrap_column = wrap_column
        self.condense = condense

        if self.wrap_column:
            # For bullets, we need two text wrappers: one for the leading
            # bullet on the first paragraph, one without.
            self.bullet_tw1 = textwrap.TextWrapper(
                width=self.wrap_column,
                initial_indent=self.BULLET_INDENT1,
                subsequent_indent=self.BULLET_INDENT2)

            self.bullet_tw2 = textwrap.TextWrapper(
                width=self.wrap_column,
                initial_indent=self.BULLET_INDENT2,
                subsequent_indent=self.BULLET_INDENT2)

            # For blockquotes, each line is prefixed with "> "
            self.quote_tw = textwrap.TextWrapper(
                width=self.wrap_column,
                initial_indent=self.QUOTE_INDENT,
                subsequent_indent=self.QUOTE_INDENT)

    def print_file(
        self,
        filename: str,
        document: Document
    ) -> typ.Iterator[str]:
        body_iter = self.emit_body(document)

        if self.print_filename:
            # Print the file name, only if there is some output.
            try:
                first = next(body_iter)
            except StopIteration:
                pass
            else:
                yield "# File: '%s'\n\n" % filename
                yield first

        yield from body_iter

    @staticmethod
    def format_pos(
        pos: Pos,
        document: Document,
        use_page_label: bool,
        page_number_offset: int
    ) -> str:

        result = pos.page.format_name(
            use_label=use_page_label,
            page_number_offset=page_number_offset).title()

        o = document.nearest_outline(pos)
        if o:
            result += " (%s)" % o.title

        return result

    def format_bullet(
        self,
        paras: typ.List[str],
        quote: typ.Optional[typ.Tuple[int, int]] = None
    ) -> str:
        """
        Format a Markdown bullet, wrapped as desired.
        """

        if quote is not None:
            (quotepos, quotelen) = quote
            assert quotepos > 0  # first paragraph to format as a block-quote
            assert quotelen > 0  # length of the blockquote in paragraphs
            assert quotepos + quotelen <= len(paras)

        # emit the first paragraph with the bullet
        if self.wrap_column:
            ret = self.bullet_tw1.fill(paras[0])
        else:
            ret = self.BULLET_INDENT1 + paras[0]

        # emit subsequent paragraphs
        npara = 1
        for para in paras[1:]:
            # are we in a blockquote?
            inquote = quote and npara >= quotepos and npara < quotepos + quotelen

            # emit a paragraph break
            # if we're going straight to a quote, we don't need an extra newline
            ret = ret + ('\n' if quote and npara == quotepos else '\n\n')

            if self.wrap_column:
                tw = self.quote_tw if inquote else self.bullet_tw2
                ret = ret + tw.fill(para)
            else:
                indent = self.QUOTE_INDENT if inquote else self.BULLET_INDENT2
                ret = ret + indent + para

            npara += 1

        return ret

    def merge_context(self, annot: Annotation, text: str) -> str:
        """Merge the context for a strikeout or caret annotation into the text."""
        (pre, post) = annot.get_context(self.remove_hyphens)

        if pre:
            pre = trim_context(pre, keep_right=True)

        if post:
            post = trim_context(post, keep_right=False)

        if annot.subtype == AnnotationType.StrikeOut:
            return pre + '~~' + text + '~~' + post
        else:
            assert annot.subtype == AnnotationType.Caret
            assert text.isspace()
            return pre.rstrip(' ') + ' ^ ' + post.lstrip(' ')

    def format_annot(
        self,
        annot: Annotation,
        document: Document,
        extra: typ.Optional[str] = None
    ) -> str:
        # Limited support for Caret annotations with a single "reply" of type StrikeOut
        contents = annot.contents
        if annot.subtype == AnnotationType.Caret and annot.group_children:
            child = annot.get_child_by_type(AnnotationType.StrikeOut)
            if child:
                annot = child
                if child.contents:
                    logger.warning("Ignored StrikeOut comment: %s", child.contents)

        # capture item text and contents (i.e. the comment), and split the latter into paragraphs
        text = annot.gettext(self.remove_hyphens) or ''
        comment = [l for l in contents.splitlines() if l] if contents else []

        if annot.has_context():
            text = self.merge_context(annot, text)

        # we are either printing: item text and item contents, or one of the two
        # if we see an annotation with neither, something has gone wrong
        if not (text or comment):
            logger.warning('%s annotation at %s has neither text nor a comment; skipped',
                           annot.subtype.name, annot.pos)
            return ''

        # compute the formatted position (and extra bit if needed) as a label
        assert annot.pos is not None
        label = self.format_pos(
            annot.pos, document, self.use_page_labels, self.page_number_offset
        ) + (" " + extra if extra else "") + ":"

        # If we have short (few words) text with a short or no comment, and the
        # text contains no embedded full stops or quotes, then we'll just put
        # quotation marks around the text and merge the two into a single paragraph.
        if (self.condense
            and text
            and not annot.has_context()
            and len(text.split()) <= 10  # words
            and all([x not in text for x in ['"', '. ']])
                and (not comment or len(comment) == 1)):
            msg = label + ' "' + text + '"'
            if comment:
                msg = msg + ' -- ' + comment[0]
            return self.format_bullet([msg]) + "\n\n"

        # If there is no text and a single-paragraph comment, it also goes on
        # one line.
        elif comment and not text and len(comment) == 1:
            msg = label + " " + comment[0]
            return self.format_bullet([msg]) + "\n\n"

        # Otherwise, text (if any) turns into a blockquote, and the comment (if
        # any) into subsequent paragraphs.
        else:
            msgparas = [label] + [text] + comment
            quotepos = (1, 1) if text else None
            return self.format_bullet(msgparas, quotepos) + "\n\n"

    def emit_body(
        self,
        document: Document
    ) -> typ.Iterator[str]:
        for a in document.iter_annots():
            yield self.format_annot(a, document, a.subtype.name)


class GroupedMarkdownPrinter(MarkdownPrinter):
    ANNOT_NITS = frozenset({AnnotationType.Caret, AnnotationType.Squiggly,
                            AnnotationType.StrikeOut, AnnotationType.Underline})
    ALL_SECTIONS = ["highlights", "comments", "nits"]

    def __init__(
        self,
        *,
        sections: typ.Sequence[str] = ALL_SECTIONS,  # controls the order of sections output
        group_highlights_by_color: bool = False,     # Whether to group highlights by color
        **kwargs: typ.Any                            # other args -- see superclass
    ) -> None:
        super().__init__(**kwargs)
        self.sections = sections
        self.group_highlights_by_color = group_highlights_by_color
        self._fmt_header_called: bool

    def emit_body(
        self,
        document: Document
    ) -> typ.Iterator[str]:

        self._fmt_header_called = False

        def fmt_header(name: str, level: int = 2) -> str:
            """
            A function that formats a header with a given name and level.

            Parameters:
                name (str): The name of the header.
                level (int, optional): The level of the header. Defaults to 2.

            Returns:
                str: The formatted header.
            """
            # emit blank separator line if needed
            prefix = '\n' if self._fmt_header_called else ''
            self._fmt_header_called = True
            header = '#' * level
            return prefix + header + " " + name + "\n"

        # Partition annotations into nits, comments, and highlights.
        nits: typ.List[Annotation] = []
        comments: typ.List[Annotation] = []
        highlights: typ.List[Annotation] = []  # When grouping by color holds only undefined annots
        highlights_by_color: typ.DefaultDict[RGB, typ.List[Annotation]] = defaultdict(list)

        for a in document.iter_annots():
            if a.subtype in self.ANNOT_NITS:
                nits.append(a)
            elif a.contents:
                comments.append(a)
            elif a.subtype == AnnotationType.Highlight:
                if self.group_highlights_by_color and a.color:
                    highlights_by_color[a.color].append(a)
                else:
                    highlights.append(a)

        for secname in self.sections:
            if (highlights or highlights_by_color) and secname == 'highlights':
                yield fmt_header("Highlights")

                for color, annots in highlights_by_color.items():
                    yield fmt_header(f"Color: {color.ashex()}", level=3)
                    for a in annots:
                        yield self.format_annot(a, document)

                if highlights and self.group_highlights_by_color:
                    yield fmt_header("Color: undefined", level=3)

                for a in highlights:
                    yield self.format_annot(a, document)

            if comments and secname == 'comments':
                yield fmt_header("Detailed comments")
                for a in comments:
                    yield self.format_annot(a, document)

            if nits and secname == 'nits':
                yield fmt_header("Nits")
                for a in nits:
                    extra = None
                    if a.subtype == AnnotationType.Caret:
                        if a.get_child_by_type(AnnotationType.StrikeOut):
                            extra = "suggested replacement"
                        else:
                            extra = "suggested insertion"
                    elif a.subtype == AnnotationType.StrikeOut:
                        extra = "suggested deletion"

                    yield self.format_annot(a, document, extra)



import abc
import typing as typ

from ..types import Document


class Printer(abc.ABC):
    """
    Base class for pretty-printers.
    """

    def begin(self) -> str:
        """Called once prior to print_file call. Returns initial output."""
        return ''

    @abc.abstractmethod
    def print_file(
        self,
        filename: str,
        document: Document
    ) -> typ.Iterator[str]:
        """
        Pretty-print a single document.

        Pretty-print the extracted annotations, yielding output (incrementally) as strings.
        Called multiple times, once per file.
        """

    def end(self) -> str:
        """Called once after the final print_file call. Returns any final additional output."""
        return ''



import datetime
import typing as typ

CHARACTER_SUBSTITUTIONS = {
    'ﬀ': 'ff',
    'ﬁ': 'fi',
    'ﬂ': 'fl',
    'ﬃ': 'ffi',
    'ﬄ': 'ffl',
    '‘': "'",
    '’': "'",
    '“': '"',
    '”': '"',
    '…': '...',
}


def cleanup_text(text: str) -> str:
    """
    Normalise line endings and replace common special characters with plain ASCII equivalents.
    """
    if '\r' in text:
        text = text.replace('\r\n', '\n').replace('\r', '\n')
    return ''.join([CHARACTER_SUBSTITUTIONS.get(c, c) for c in text])


def merge_lines(captured_text: str, remove_hyphens: bool = False, strip_space: bool = True) -> str:
    """
    Merge and cleanup lines in captured text, optionally removing hyphens.

    Any number of consecutive newlines is replaced by a single space, unless the
    prior line ends in a hyphen, in which case they are just removed entirely.
    This makes it easier for the renderer to "broadcast" newlines to active
    annotations regardless of box hits. (Detecting paragraph breaks is tricky,
    and left for future work!)
    """
    results = []

    lines = captured_text.splitlines()
    for i in range(len(lines)):
        thisline = lines[i]
        if thisline == '':
            continue

        nextline = lines[i + 1] if i + 1 < len(lines) else None

        if (len(thisline) >= 2
                and thisline[-1] == '-'       # Line ends in an apparent hyphen
                and thisline[-2].islower()):  # Prior character was a lowercase letter
            # We have a likely hyphen. Remove it if desired.
            if remove_hyphens:
                thisline = thisline[:-1]
        elif (not thisline[-1].isspace()
              and nextline is not None
              and (nextline == '' or not nextline[0].isspace())):
            # Insert space to replace the line break
            thisline += ' '

        results.append(cleanup_text(thisline))

    result = ''.join(results)

    if result:
        if strip_space:
            result = result.strip()
        else:
            # re-insert load-bearing spaces from linebreaks when needed for context
            if len(lines) > 0 and lines[0] == '' and not result[0].isspace():
                result = ' ' + result
            if len(lines) > 1 and lines[-1] == '' and not result[-1].isspace():
                result += ' '

    return result


def decode_datetime(dts: str) -> typ.Optional[datetime.datetime]:
    if dts.startswith('D:'):  # seems 'optional but recommended'
        dts = dts[2:]
    dts = dts.replace("'", '')
    zi = dts.find('Z')
    if zi != -1:  # sometimes it's Z/Z0000
        dts = dts[:zi] + '+0000'
    fmt = '%Y%m%d%H%M%S'
    # dates in PDFs are quite flaky and underspecified... so perhaps worth defensive code here
    for suf in ['%z', '']:  # sometimes timezone is missing
        try:
            return datetime.datetime.strptime(dts, fmt + suf)
        except ValueError:
            continue
    return None



import argparse
import logging
import sys
import typing as typ

from pdfminer.layout import LAParams

from . import __doc__, __version__, process_file
from .printer import Printer
from .printer.markdown import MarkdownPrinter, GroupedMarkdownPrinter
from .printer.json import JsonPrinter


MD_FORMAT_ARGS = frozenset([
    'condense',
    'group_highlights_by_color',
    'page_number_offset',
    'print_filename',
    'sections',
    'use_page_labels',
    'wrap_column',
])
"""Names of arguments passed to the markdown printer."""


def _float_or_disabled(x: str) -> typ.Optional[float]:
    if x.lower().strip() == "disabled":
        return None
    try:
        return float(x)
    except ValueError as ex:
        raise argparse.ArgumentTypeError("invalid float value: {}".format(x)) from ex


def parse_args() -> typ.Tuple[argparse.Namespace, LAParams]:
    p = argparse.ArgumentParser(prog='pdfannots', description=__doc__)

    p.add_argument('--version', action='version',
                   version='%(prog)s ' + __version__)

    p.add_argument("input", metavar="INFILE", type=argparse.FileType("rb"),
                   help="PDF files to process", nargs='+')

    g = p.add_argument_group('Basic options')
    g.add_argument("-p", "--progress", default=False, action="store_true",
                   help="Emit progress information to stderr.")
    g.add_argument("-o", metavar="OUTFILE", type=argparse.FileType("w", encoding="utf-8"),
                   dest="output", default=sys.stdout, help="Output file (default is stdout).")
    g.add_argument("-n", "--cols", default=None, type=int, metavar="COLS", dest="cols",
                   help="Assume a fixed top-to-bottom left-to-right page layout with this many "
                        "columns per page. If unset, PDFMiner's layout detection logic is used.")
    g.add_argument("--keep-hyphens", dest="remove_hyphens", default=True, action="store_false",
                   help="When capturing text across a line break, don't attempt to remove hyphens.")
    g.add_argument("-f", "--format", choices=["md", "json"], default="md",
                   help="Output format (default: markdown).")

    g = p.add_argument_group('Options controlling markdown output')
    mutex_group = g.add_mutually_exclusive_group()
    mutex_group.add_argument(
        "--no-group",
        dest="group",
        default=True, action="store_false",
        help="Emit annotations in order, don't group into sections."
    )
    mutex_group.add_argument(
        "--group-highlights-by-color",
        dest="group_highlights_by_color",
        default=False, action="store_true",
        help="Group highlights by color in grouped output."
    )

    g.add_argument("-s", "--sections", metavar="SEC", nargs="*",
                   choices=GroupedMarkdownPrinter.ALL_SECTIONS,
                   default=GroupedMarkdownPrinter.ALL_SECTIONS,
                   help=("sections to emit (default: %s)" %
                         ', '.join(GroupedMarkdownPrinter.ALL_SECTIONS)))
    g.add_argument("--no-condense", dest="condense", default=True, action="store_false",
                   help="Emit annotations as a blockquote regardless of length.")
    g.add_argument("--no-page-labels", dest="use_page_labels", default=True, action="store_false",
                   help="Ignore page labels if present, just print 1-based page numbers.")
    g.add_argument("--page-number-offset", dest="page_number_offset", default=0, type=int,
                   help="Increase or decrease page numbers with a fixed offset.")
    g.add_argument("--print-filename", dest="print_filename", default=False, action="store_true",
                   help="Print the name of each file with annotations.")
    g.add_argument("-w", "--wrap", dest="wrap_column", metavar="COLS", type=int,
                   help="Wrap text at this many output columns.")

    g = p.add_argument_group(
        "Advanced options affecting PDFMiner text layout analysis")
    laparams = LAParams()
    g.add_argument(
        "--line-overlap", metavar="REL_HEIGHT", type=float, default=laparams.line_overlap,
        help="If two characters have more overlap than this they are considered to be "
             "on the same line. The overlap is specified relative to the minimum height "
             "of both characters. Default: %s" % laparams.line_overlap)
    g.add_argument(
        "--char-margin", metavar="REL_WIDTH", type=float, default=laparams.char_margin,
        help="If two characters are closer together than this margin they "
             "are considered to be part of the same line. The margin is "
             "specified relative to the character width. Default: %s" % laparams.char_margin)
    g.add_argument(
        "--word-margin", metavar="REL_WIDTH", type=float, default=laparams.word_margin,
        help="If two characters on the same line are further apart than this "
             "margin then they are considered to be two separate words, and "
             "an intermediate space will be added for readability. The margin "
             "is specified relative to the character width. Default: %s" % laparams.word_margin)
    g.add_argument(
        "--line-margin", metavar="REL_HEIGHT", type=float, default=laparams.line_margin,
        help="If two lines are close together they are considered to "
             "be part of the same paragraph. The margin is specified "
             "relative to the height of a line. Default: %s" % laparams.line_margin)
    g.add_argument(
        "--boxes-flow", type=_float_or_disabled, default=laparams.boxes_flow,
        help="Specifies how much a horizontal and vertical position of a "
             "text matters when determining the order of lines. The value "
             "should be within the range of -1.0 (only horizontal position "
             "matters) to +1.0 (only vertical position matters). You can also "
             "pass 'disabled' to disable advanced layout analysis, and "
             "instead return text based on the position of the bottom left "
             "corner of the text box. Default: %s" % laparams.boxes_flow)

    # The next two booleans are described as if they default off, so let's ensure that.
    assert not laparams.detect_vertical
    assert not laparams.all_texts
    g.add_argument(
        "--detect-vertical", default=laparams.detect_vertical,
        action="store_const", const=(not laparams.detect_vertical),
        help="Consider vertical text during layout analysis.")
    g.add_argument(
        "--all-texts", default=laparams.all_texts,
        action="store_const", const=(not laparams.all_texts),
        help="Perform layout analysis on text in figures.")

    args = p.parse_args()

    # Propagate parsed layout parameters back to LAParams object
    for param in ("line_overlap", "char_margin", "word_margin", "line_margin",
                  "boxes_flow", "detect_vertical", "all_texts"):
        setattr(laparams, param, getattr(args, param))

    return args, laparams


def main() -> None:
    args, laparams = parse_args()
    logging.basicConfig(format='%(levelname)s: %(message)s',
                        level=logging.WARNING)

    # construct appropriate Printer
    printer: Printer
    if args.format == "md":
        mdargs = {k: getattr(args, k) for k in MD_FORMAT_ARGS}
        printer = (GroupedMarkdownPrinter if args.group else MarkdownPrinter)(**mdargs)
    elif args.format == "json":
        printer = JsonPrinter(
            remove_hyphens=args.remove_hyphens,
            output_codec=args.output.encoding)

    def write_if_nonempty(s: str) -> None:
        if s:
            args.output.write(s)

    write_if_nonempty(printer.begin())

    # iterate over files
    for file in args.input:
        doc = process_file(
            file,
            columns_per_page=args.cols,
            emit_progress_to=(sys.stderr if args.progress else None),
            laparams=laparams)
        for line in printer.print_file(file.name, doc):
            args.output.write(line)

    write_if_nonempty(printer.end())



#!/usr/bin/env python3

import functools
import json
import operator
import pathlib
import re
import typing as typ
import unittest
from datetime import datetime, timedelta, timezone

import pdfminer.layout

import pdfannots
import pdfannots.utils
from pdfannots.types import AnnotationType
from pdfannots.printer.markdown import MarkdownPrinter, GroupedMarkdownPrinter
from pdfannots.printer.json import JsonPrinter


class UnitTests(unittest.TestCase):
    def test_decode_datetime(self) -> None:
        datas = [
            ("D:123456", None),  # defensive on bad datetimes
            ("D:20190119212926-08'00'",
             datetime(2019, 1, 19, 21, 29, 26, tzinfo=timezone(-timedelta(hours=8)))),
            ("20200102030405Z0000",
             datetime(2020, 1, 2, 3, 4, 5, tzinfo=timezone.utc)),
            ("D:20101112191817", datetime(2010, 11, 12, 19, 18, 17)),
        ]
        for dts, expected in datas:
            dt = pdfannots.utils.decode_datetime(dts)
            self.assertEqual(dt, expected)


class ExtractionTestBase(unittest.TestCase):
    filename: str

    # Permit a test to customise the columns_per_page or LAParams
    columns_per_page: typ.Optional[int] = None
    laparams = pdfminer.layout.LAParams()

    def setUp(self) -> None:
        path = pathlib.Path(__file__).parent / 'tests' / self.filename
        with path.open('rb') as f:
            self.doc = pdfannots.process_file(f, columns_per_page=self.columns_per_page,
                                              laparams=self.laparams)
            self.annots = [a for p in self.doc.pages for a in p.annots]
            self.outlines = [o for p in self.doc.pages for o in p.outlines]

    def assertEndsWith(self, bigstr: str, suffix: str) -> None:
        self.assertEqual(bigstr[-len(suffix):], suffix)

    def assertStartsWith(self, bigstr: str, prefix: str) -> None:
        self.assertEqual(bigstr[:len(prefix)], prefix)


class ExtractionTests(ExtractionTestBase):
    filename = 'hotos17.pdf'
    columns_per_page = 2  # for test_nearest_outline

    def test_annots(self) -> None:
        EXPECTED = [
            (0, AnnotationType.Squiggly, None, 'recent Intel CPUs have introduced'),
            (0, AnnotationType.Text, 'This is a note with no text attached.', None),
            (0, AnnotationType.StrikeOut, None, 'e'),
            (1, AnnotationType.Highlight, None,
             'TSX launched with "Haswell" in 2013 but was later disabled due to a bug. '
             '"Broadwell" CPUs with the bug fix shipped in late 2014.'),
            (1, AnnotationType.Highlight, 'This is lower in column 1',
             'user-mode access to FS/GS registers, and TLB tags for non-VM address spaces'),
            (1, AnnotationType.Highlight, None,
             'segmentation, task switching, and 16-bit modes.'),
            (1, AnnotationType.Highlight, 'This is at the top of column two',
             'The jump is due to extensions introduced with the "Skylake" microarchitecture'),
            (3, AnnotationType.Squiggly, 'This is a nit.',
             'Control transfer in x86 is already very complex'),
            (3, AnnotationType.Underline, 'This is a different nit',
             'Besides modifying semantics of all indirect control transfers'),
            (3, AnnotationType.StrikeOut, None,
             'While we may disagree with some of the design choices,')]

        self.assertEqual(len(self.annots), len(EXPECTED))
        for a, expected in zip(self.annots, EXPECTED):
            assert a.pos is not None
            self.assertEqual(
                (a.pos.page.pageno, a.subtype, a.contents, a.gettext(remove_hyphens=True)),
                expected)
        self.assertEqual(self.annots[0].created, datetime(
            2019, 1, 19, 21, 29, 42, tzinfo=timezone(-timedelta(hours=8))))

        # test for correct whitespace on the strikeout annot
        a = self.annots[2]
        self.assertTrue(a.has_context())
        (pre, post) = a.get_context()
        self.assertEndsWith(pre, 'widths, ar')
        self.assertStartsWith(post, ' counted')

    def test_outlines(self) -> None:
        EXPECTED = [
            'Introduction',
            'Background: x86 extensions',
            'Case study: SGX',
            'Case study: CET',
            'Implications',
            'Concluding remarks']

        self.assertEqual(len(self.outlines), len(EXPECTED))
        for o, expected in zip(self.outlines, EXPECTED):
            self.assertEqual(o.title, expected)

    def test_nearest_outline(self) -> None:
        # Page 1 (Introduction) Squiggly: "recent Intel CPUs have introduced"
        a = self.doc.pages[0].annots[0]
        assert a.pos is not None
        o = self.doc.nearest_outline(a.pos)
        assert o is not None
        self.assertEqual(o.title, 'Introduction')

        # Page 4 (Case study: CET) Squiggly: "Control transfer in x86 is already very complex"
        # Note: pdfminer gets this wrong as of 20201018; we must set columns_per_page to fix it
        a = self.doc.pages[3].annots[0]
        assert a.pos is not None
        o = self.doc.nearest_outline(a.pos)
        assert o is not None
        self.assertEqual(o.title, 'Case study: CET')


class Issue9(ExtractionTestBase):
    filename = 'issue9.pdf'

    def test(self) -> None:
        self.assertEqual(len(self.annots), 1)
        a = self.annots[0]
        self.assertEqual(a.gettext(), 'World')


class Issue13(ExtractionTestBase):
    filename = 'issue13.pdf'

    def test(self) -> None:
        self.assertEqual(len(self.annots), 1)
        a = self.annots[0]
        self.assertEqual(a.gettext(), 'This is a sample statement.')


class Issue46(ExtractionTestBase):
    filename = 'issue46.pdf'

    def test(self) -> None:
        self.assertEqual(len(self.annots), 3)

        self.assertEqual(self.annots[0].subtype, AnnotationType.Highlight)
        self.assertEqual(self.annots[0].gettext(), 'C – Curate')

        self.assertEqual(self.annots[1].subtype, AnnotationType.Square)
        self.assertEqual(self.annots[1].gettext(), None)

        self.assertEqual(self.annots[2].subtype, AnnotationType.Highlight)
        self.assertEqual(self.annots[2].gettext(), 'This was a novel idea at the time')


class Issue61(ExtractionTestBase):
    filename = 'issue61.pdf'

    def test(self) -> None:
        self.assertEqual(len(self.annots), 1)
        a = self.annots[0]
        self.assertEqual(a.subtype, AnnotationType.Caret)
        self.assertEqual(a.contents, 'and machine learning')
        self.assertTrue(a.has_context())


class Pr24(ExtractionTestBase):
    filename = 'pr24.pdf'

    def test(self) -> None:
        EXPECTED = [
            (AnnotationType.Highlight, 'long highlight',
             'Heading Link to heading that is working with vim-pandoc. Link to heading that'),
            (AnnotationType.Highlight, 'short highlight', 'not working'),
            (AnnotationType.Text, None, None),
            (AnnotationType.Highlight, None, 'Some more text'),
            (AnnotationType.Text, 'dual\n\npara note', None),
            (AnnotationType.Text, 's', None)]
        self.assertEqual(len(self.annots), len(EXPECTED))
        for a, expected in zip(self.annots, EXPECTED):
            self.assertEqual((a.subtype, a.contents, a.gettext()), expected)


class Landscape2Column(ExtractionTestBase):
    filename = 'word2column.pdf'

    def test(self) -> None:
        self.assertEqual(len(self.annots), 9)

        a = self.annots[0]
        self.assertEqual(a.subtype, AnnotationType.StrikeOut)
        self.assertEqual(a.gettext(), 'nostrud exercitation')
        self.assertTrue(a.has_context())
        (pre, post) = a.get_context()
        self.assertEndsWith(
            pre, 'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor '
            'incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis ')
        self.assertStartsWith(
            post, ' ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor')

        a = self.annots[1]
        self.assertEqual(a.subtype, AnnotationType.StrikeOut)
        self.assertEqual(a.gettext(), 'Duis')
        self.assertTrue(a.has_context())
        (pre, post) = a.get_context()
        self.assertEndsWith(pre, 'ullamco laboris nisi ut aliquip ex ea commodo consequat. ')
        self.assertStartsWith(
            post, ' aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu '
            'fugiat nulla pariatur.')

        a = self.annots[2]
        self.assertEqual(a.subtype, AnnotationType.StrikeOut)
        self.assertEqual(a.gettext(), 'laborum')
        self.assertTrue(a.has_context())
        (pre, post) = a.get_context()
        self.assertEndsWith(pre, ', sunt in culpa qui officia deserunt mollit anim id est ')
        self.assertStartsWith(post, '. Heading 2 Sed ut perspiciatis,')

        a = self.annots[3]
        self.assertEqual(a.subtype, AnnotationType.Highlight)
        self.assertEqual(
            a.gettext(), 'At vero eos et accusamus et iusto odio dignissimos ducimus, qui '
            'blanditiis praesentium voluptatum deleniti atque corrupti,')
        self.assertFalse(a.has_context())

        a = self.annots[4]
        self.assertEqual(a.subtype, AnnotationType.Squiggly)
        self.assertEqual(
            a.gettext(), 'Itaque earum rerum hic tenetur a sapiente delectus, ut aut reiciendis '
            'voluptatibus maiores alias consequatur aut perferendis doloribus asperiores repellat.')
        self.assertEqual(a.contents, 'Nonsense!')
        self.assertFalse(a.has_context())

        a = self.annots[5]
        self.assertEqual(a.subtype, AnnotationType.StrikeOut)
        self.assertEqual(a.gettext(), 'equal')
        self.assertTrue(a.has_context())
        (pre, post) = a.get_context()
        self.assertEndsWith(pre, 'the pain and trouble that are bound to ensue; and ')
        self.assertStartsWith(post, ' blame belongs to those who fail in their')  # end of page

        a = self.annots[6]
        self.assertEqual(a.subtype, AnnotationType.StrikeOut)
        self.assertEqual(a.gettext(), 'duty')
        self.assertTrue(a.has_context())
        (pre, post) = a.get_context()
        self.assertEqual(pre, '')  # start of page
        self.assertStartsWith(post, ' through weakness of will, which')

        a = self.annots[7]
        self.assertEqual(a.subtype, AnnotationType.StrikeOut)
        self.assertEqual(a.gettext(), 'In a free hour,')
        self.assertTrue(a.has_context())
        (pre, post) = a.get_context()
        self.assertEndsWith(pre, 'These cases are perfectly simple and easy to distinguish. ')
        self.assertStartsWith(post, ' when our power of choice is untrammeled and when nothing')


class FreeTextAnnotation(ExtractionTestBase):
    filename = 'FreeText-annotation.pdf'

    def test(self) -> None:
        self.assertEqual(len(self.annots), 1)
        self.assertEqual(self.annots[0].subtype, AnnotationType.FreeText)
        self.assertEqual(self.annots[0].contents, 'Annotation with subtype "FreeText".')
        self.assertEqual(self.annots[0].gettext(), None)


class CaretAnnotations(ExtractionTestBase):
    filename = 'caret.pdf'

    def test(self) -> None:
        self.assertEqual(len(self.annots), 5)
        a = self.annots[0]
        self.assertEqual(a.subtype, AnnotationType.StrikeOut)
        self.assertEqual(a.gettext(), 'Adobe Acrobat Reader')
        self.assertTrue(a.is_group_child)
        self.assertEqual(a.group_children, [])
        g = self.annots[3]
        self.assertEqual(g.subtype, AnnotationType.Caret)
        self.assertEqual(g.contents, 'Google Chrome')
        self.assertFalse(g.is_group_child)
        self.assertEqual(g.group_children, [a])
        self.assertEqual(g.get_child_by_type(AnnotationType.StrikeOut), a)


class PrinterTestBase(unittest.TestCase):
    filename = 'hotos17.pdf'

    def setUp(self) -> None:
        path = pathlib.Path(__file__).parent / 'tests' / self.filename
        with path.open('rb') as f:
            self.doc = pdfannots.process_file(f)


class MarkdownPrinterTest(PrinterTestBase):
    # There's not a whole lot of value in testing the precise output format,
    # but let's make sure we produce a non-trivial result and don't crash.
    def test_flat(self) -> None:
        p = MarkdownPrinter(print_filename=True, remove_hyphens=False)

        linecount = 0
        charcount = 0
        for line in p.print_file('dummyfile', self.doc):
            linecount += line.count('\n')
            charcount += len(line)

        self.assertGreater(linecount, 5)
        self.assertGreater(charcount, 500)

    def test_flat_page_number_offset(self) -> None:
        p = MarkdownPrinter(page_number_offset=-1)

        page_numbers = []
        for line in p.print_file('dummyfile', self.doc):
            m = re.match(r'.+Page #([0-9])', line)
            if m:
                page_numbers.append(m[1])

        self.assertEqual(page_numbers, ['0', '0', '0', '1', '1', '1', '1', '3', '3', '3'])

    def test_grouped(self) -> None:
        p = GroupedMarkdownPrinter(wrap_column=80)

        linecount = 0
        charcount = 0
        for line in p.print_file('dummyfile', self.doc):
            linecount += line.count('\n')
            charcount += len(line)

        self.assertGreater(linecount, 10)
        self.assertGreater(charcount, 900)

    def test_multicolorgrouping(self) -> None:
        p = GroupedMarkdownPrinter(group_highlights_by_color=True)

        linecount = 0
        charcount = 0
        for line in p.print_file('dummyfile', self.doc):
            linecount += line.count('\n')
            charcount += len(line)

        self.assertGreater(linecount, 10)
        self.assertGreater(charcount, 900)


class JsonPrinterTest(PrinterTestBase):
    def test_flat(self) -> None:
        p = JsonPrinter(remove_hyphens=False, output_codec='utf-8')

        j = json.loads(
            p.begin()
            + functools.reduce(operator.add, p.print_file('dummyfile', self.doc))
            + p.end())

        self.assertTrue(isinstance(j, list))
        self.assertEqual(len(j), 10)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import bisect
import datetime
import enum
import functools
import logging
import typing as typ

from pdfminer.layout import LTComponent, LTText
from pdfminer.pdftypes import PDFObjRef

from .utils import merge_lines

logger = logging.getLogger('pdfannots')

Point = typ.Tuple[float, float]
"""An (x, y) point in PDF coordinates, i.e. bottom left is 0,0."""

BoxCoords = typ.Tuple[float, float, float, float]
"""The coordinates of a bounding box (x0, y0, x1, y1)."""


class Box:
    """
    Coordinates of a rectangular box.
    """

    def __init__(self, x0: float, y0: float, x1: float, y1: float):
        assert x0 <= x1 and y0 <= y1
        self.x0 = x0
        self.x1 = x1
        self.y0 = y0
        self.y1 = y1

    def __repr__(self) -> str:
        return '<Box (%f,%f) (%f,%f)>' % (self.x0, self.y0, self.x1, self.y1)

    @staticmethod
    def from_item(item: LTComponent) -> Box:
        """Construct a Box from the bounding box of a given PDF component."""
        return Box(item.x0, item.y0, item.x1, item.y1)

    @staticmethod
    def from_coords(coords: BoxCoords) -> Box:
        """Construct a Box from the given PDF coordinates."""
        (x0, y0, x1, y1) = coords
        return Box(x0, y0, x1, y1)

    def get_coords(self) -> BoxCoords:
        """Return the PDF coordinates of this box."""
        return (self.x0, self.y0, self.x1, self.y1)

    def get_width(self) -> float:
        """Return the width of the box."""
        return self.x1 - self.x0

    def get_height(self) -> float:
        """Return the height of the box."""
        return self.y1 - self.y0

    def get_area(self) -> float:
        """Return the area of the box."""
        return self.get_height() * self.get_width()

    def get_overlap(self, other: Box) -> float:
        """Compute the overlapping area (if any) with the provided box."""
        x_overlap = max(0, min(other.x1, self.x1) - max(other.x0, self.x0))
        y_overlap = max(0, min(other.y1, self.y1) - max(other.y0, self.y0))
        return x_overlap * y_overlap

    def hit_item(self, item: LTComponent) -> bool:
        """Does most of the area of the PDF component overlap this box?"""
        item_area = float(item.width) * float(item.height)
        overlap_area = self.get_overlap(Box.from_item(item))

        if overlap_area != 0:
            logger.debug(
                "Box hit: '%s' %f-%f,%f-%f in %f-%f,%f-%f %2.0f%%",
                item.get_text() if isinstance(item, LTText) else '',
                item.x0, item.x1, item.y0, item.y1,
                self.x0, self.x1, self.y0, self.y1,
                100 * overlap_area / item_area)

        assert overlap_area <= item_area
        return (item_area != 0) and overlap_area >= (0.5 * item_area)

    def closest_point(self, point: Point) -> Point:
        """Compute the closest point in this box to the specified point."""
        px, py = point
        return (min(max(self.x0, px), self.x1),
                min(max(self.y0, py), self.y1))

    def square_of_distance_to_closest_point(self, point: Point) -> float:
        """
        Compute the distance from the closest point in this box to the specified point, squared.

        (We avoid calling sqrt for performance reasons, since we just need to compare.)
        """
        x, y = self.closest_point(point)
        px, py = point
        return abs(px - x)**2 + abs(py - y)**2


@functools.total_ordering
class Page:
    """
    Page.

    A page object uniquely represents a page in the PDF. It is identified by a
    zero-based page number, and a PDF object ID. It holds a list of Annotation
    objects for annotations on the page, and Outline objects for outlines that
    link to somewhere on the page.
    """

    annots: typ.List[Annotation]
    outlines: typ.List[Outline]

    def __init__(
        self,
        pageno: int,
        objid: object,
        label: typ.Optional[str],
        mediabox: BoxCoords,
        fixed_columns: typ.Optional[int] = None
    ):
        assert pageno >= 0
        assert fixed_columns is None or fixed_columns > 0
        self.pageno = pageno
        self.objid = objid
        self.label = label
        self.annots = []
        self.outlines = []
        self.mediabox = Box.from_coords(mediabox)
        self.fixed_columns = fixed_columns

    def __repr__(self) -> str:
        return '<Page #%d>' % self.pageno  # zero-based page index

    def __str__(self) -> str:
        return self.format_name()

    def format_name(self, use_label: bool = True, page_number_offset: int = 0) -> str:
        if self.label and use_label:
            return 'page %s' % self.label
        else:
            # + 1 for 1-based page numbers in normal program output (error messages, etc.)
            return 'page #%d' % (self.pageno + 1 + page_number_offset)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Page):
            return NotImplemented
        return self.pageno == other.pageno

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Page):
            return NotImplemented
        return self.pageno < other.pageno


@functools.total_ordering
class Pos:
    """
    A position within the document.

    This object represents an x,y point on a particular page. Such positions are
    also comparable, and compare in natural document reading order (as inferred
    by pdfminer's text layout detection).
    """

    def __init__(self, page: Page, x: float, y: float):
        self.page = page
        self.x = x
        self.y = y
        self._pageseq = 0
        self._pageseq_distance = 0.0

    def __str__(self) -> str:
        return '%s (%.3f,%.3f)' % (self.page, self.x, self.y)

    def __repr__(self) -> str:
        return '<Pos pg#%d (%.3f,%.3f) #%d>' % (self.page.pageno, self.x, self.y, self._pageseq)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Pos):
            return (self.page == other.page
                    and self.x == other.x
                    and self.y == other.y)
        return NotImplemented

    def __lt__(self, other: object) -> bool:
        if isinstance(other, Pos):
            if self.page == other.page:
                assert self.page is other.page
                if self.page.fixed_columns:
                    # Fixed layout: assume left-to-right top-to-bottom documents
                    (sx, sy) = self.page.mediabox.closest_point((self.x, self.y))
                    (ox, oy) = self.page.mediabox.closest_point((other.x, other.y))
                    colwidth = self.page.mediabox.get_width() / self.page.fixed_columns
                    self_col = (sx - self.page.mediabox.x0) // colwidth
                    other_col = (ox - self.page.mediabox.x0) // colwidth
                    return self_col < other_col or (self_col == other_col and sy > oy)
                else:
                    # Default layout inferred from pdfminer traversal
                    assert self._pageseq != 0
                    assert other._pageseq != 0
                    if self._pageseq == other._pageseq:
                        # The positions are on or closest to the same line of text.
                        # XXX: assume top-to-bottom left-to-right order
                        return self.x < other.x if self.y == other.y else self.y > other.y
                    else:
                        return self._pageseq < other._pageseq
            else:
                return self.page < other.page
        else:
            return NotImplemented

    def item_hit(self, item: LTComponent) -> bool:
        """Is this pos within the bounding box of the given PDF component?"""
        return (self.x >= item.x0
                and self.x <= item.x1
                and self.y >= item.y0
                and self.y <= item.y1)

    def update_pageseq(self, component: LTComponent, pageseq: int) -> bool:
        """If close-enough to the given component, adopt its sequence number and return True."""
        assert pageseq > 0
        if self.item_hit(component):
            # This pos is inside the component area
            self._pageseq = pageseq
            self._pageseq_distance = 0
            return True
        else:
            d = Box.from_item(component).square_of_distance_to_closest_point((self.x, self.y))
            if self._pageseq == 0 or self._pageseq_distance > d:
                self._pageseq = pageseq
                self._pageseq_distance = d
                return True
            return False

    def discard_pageseq(self, pageseq: int) -> None:
        """If we have been assigned the specified pageseq, forget about it."""
        if self._pageseq == pageseq:
            self._pageseq = 0
            self._pageseq_distance = 0.0


@functools.total_ordering
class ObjectWithPos:
    """Any object that (eventually) has a logical position on the page."""

    def __init__(self, pos: typ.Optional[Pos] = None):
        self.pos = pos

    def __lt__(self, other: object) -> bool:
        if isinstance(other, ObjectWithPos):
            assert self.pos is not None
            assert other.pos is not None
            return self.pos < other.pos
        return NotImplemented

    def update_pageseq(self, component: LTComponent, pageseq: int) -> bool:
        """Delegates to Pos.update_pageseq"""
        return False if self.pos is None else self.pos.update_pageseq(component, pageseq)

    def discard_pageseq(self, pageseq: int) -> None:
        """Delegates to Pos.discard_pageseq"""
        if self.pos is not None:
            self.pos.discard_pageseq(pageseq)


class AnnotationType(enum.Enum):
    """A supported PDF annotation type. Enumerant names match the Subtype names of the PDF spec."""

    # A "sticky note" comment annotation.
    Text = enum.auto()

    # Markup annotations that apply to one or more regions on the page.
    Highlight = enum.auto()
    Squiggly = enum.auto()
    StrikeOut = enum.auto()
    Underline = enum.auto()

    Caret = enum.auto()

    # A single rectangle, that is abused by some Apple tools to render custom
    # highlights. We do not attempt to capture the affected text.
    Square = enum.auto()

    # Free-form text written somewhere on the page.
    FreeText = enum.auto()


class Annotation(ObjectWithPos):
    """
    A PDF annotation, and its extracted text.

    Attributes:
        author          Author of the annotation
        color           RGB color of the annotation
        contents        Contents of the annotation in the PDF (e.g. comment/description)
        created         Timestamp the annotation was created
        group_children  Annotations grouped together with this one
        in_reply_to     Reference to another annotation on the page that this is "in reply to"
        is_group_child  Is this annotation a member of a parent group?
        last_charseq    Sequence number of the most recent character in text
        name            If present, uniquely identifies this annotation among others on the page
        replies         Annotations replying to this one (reverse of in_reply_to)
        subtype         PDF annotation type
        text            Text in the order captured (use gettext() for a cleaner form)

    Attributes updated for StrikeOut and Caret annotations:
        pre_context  Text captured just prior to the beginning of 'text'
        post_context Text captured just after the end of 'text'
    """

    boxes: typ.List[Box]
    contents: typ.Optional[str]
    group_children: typ.List[Annotation]
    in_reply_to: typ.Optional[Annotation]
    pre_context: typ.Optional[str]
    post_context: typ.Optional[str]
    replies: typ.List[Annotation]
    text: typ.List[str]

    def __init__(
            self,
            page: Page,
            subtype: AnnotationType,
            *,
            author: typ.Optional[str] = None,
            created: typ.Optional[datetime.datetime] = None,
            color: typ.Optional[RGB] = None,
            contents: typ.Optional[str] = None,
            in_reply_to_ref: typ.Optional[PDFObjRef] = None,
            is_group_child: bool = False,
            name: typ.Optional[str] = None,
            quadpoints: typ.Optional[typ.Sequence[float]] = None,
            rect: typ.Optional[BoxCoords] = None):

        # Construct boxes from quadpoints
        boxes = []
        if quadpoints is not None:
            assert len(quadpoints) % 8 == 0
            while quadpoints != []:
                (x0, y0, x1, y1, x2, y2, x3, y3) = quadpoints[:8]
                quadpoints = quadpoints[8:]
                xvals = [x0, x1, x2, x3]
                yvals = [y0, y1, y2, y3]
                box = Box(min(xvals), min(yvals), max(xvals), max(yvals))
                boxes.append(box)

        # Kludge for Caret annotations that lack quadpoints, but need to capture context
        if quadpoints is None and subtype == AnnotationType.Caret:
            assert rect is not None
            boxes.append(Box.from_coords(rect))

        # Compute a meaningful position of this annotation on the page
        assert rect or boxes
        (x0, y0, x1, y1) = rect if rect else boxes[0].get_coords()
        # XXX: assume left-to-right top-to-bottom text
        pos = Pos(page, min(x0, x1), max(y0, y1))
        super().__init__(pos)

        # Initialise the attributes
        self.author = author
        self.boxes = boxes
        self.color = color
        self.contents = contents if contents else None
        self.created = created
        self.group_children = []
        self.name = name
        self.last_charseq = 0
        self.post_context = None
        self.pre_context = None
        self.replies = []
        self.subtype = subtype
        self.text = []

        # The in_reply_to reference will be resolved in postprocess()
        self.in_reply_to = None
        self._in_reply_to_ref = in_reply_to_ref
        self.is_group_child = is_group_child
        if is_group_child:
            assert in_reply_to_ref

    def __repr__(self) -> str:
        return ('<Annotation %s %r%s%s>' %
                (self.subtype.name, self.pos,
                 " '%s'" % self.contents[:10] if self.contents else '',
                 " '%s'" % ''.join(self.text[:10]) if self.text else ''))

    def capture(self, text: str, charseq: int = 0) -> None:
        """Capture text (while rendering the PDF page)."""
        self.text.append(text)
        if charseq:
            assert charseq > self.last_charseq
            self.last_charseq = charseq

    def gettext(self, remove_hyphens: bool = False) -> typ.Optional[str]:
        """Retrieve cleaned-up text, after rendering."""
        if self.boxes:
            if self.text:
                captured = ''.join(self.text)
                return merge_lines(captured, remove_hyphens, strip_space=(not self.has_context()))
            else:
                # something's strange -- we have boxes but no text for them
                logger.warning('Missing text for %s annotation at %s', self.subtype.name, self.pos)
                return ""
        else:
            return None

    def get_child_by_type(self, child_type: AnnotationType) -> typ.Optional[Annotation]:
        """Return the first child of the given type."""
        for c in self.group_children:
            if c.subtype == child_type:
                return c
        return None

    def wants_context(self) -> bool:
        """Returns true if this annotation type should include context."""
        return self.subtype in {AnnotationType.Caret, AnnotationType.StrikeOut}

    def set_pre_context(self, pre_context: str) -> None:
        assert self.pre_context is None
        self.pre_context = pre_context

    def set_post_context(self, post_context: str) -> None:
        assert self.post_context is None

        # If the text ends in a (broadcast) newline, discard it lest it mess up the context below.
        if self.text and self.text[-1] == '\n':
            self.text.pop()

        # If the captured text ends in any (other) space, move it to the context.
        whitespace = []
        while self.text and self.text[-1].isspace():
            whitespace.append(self.text.pop())
        if whitespace:
            post_context = ''.join(whitespace) + post_context

        self.post_context = post_context

    def has_context(self) -> bool:
        """Returns true if this annotation captured context."""
        return self.pre_context is not None or self.post_context is not None

    def get_context(self, remove_hyphens: bool = False) -> typ.Tuple[str, str]:
        """Returns context captured for this annotation, as a tuple (pre, post)."""
        return (merge_lines(self.pre_context or '', remove_hyphens, strip_space=False),
                merge_lines(self.post_context or '', remove_hyphens, strip_space=False))

    def postprocess(self, annots_by_objid: typ.Dict[int, Annotation]) -> None:
        """Update internal state once all text and context has been captured."""
        # Resole the in_reply_to object reference to its annotation
        if self._in_reply_to_ref is not None:
            assert self.in_reply_to is None  # This should be called once only
            a = annots_by_objid.get(self._in_reply_to_ref.objid)
            if a is None:
                logger.warning("IRT reference (%d) not found in page annotations",
                               self._in_reply_to_ref.objid)
            elif self.is_group_child:
                a.group_children.append(self)
            else:
                self.in_reply_to = a
                a.replies.append(self)

        # The Skim PDF reader (https://skim-app.sourceforge.io/) creates annotations whose
        # default initial contents are a copy of the selected text. Unless the user goes to
        # the trouble of editing each annotation, this goes badly for us because we have
        # duplicate text and contents (e.g., for simple highlights and strikeout).
        if self.contents and (text := self.gettext()) and text.strip() == self.contents.strip():
            self.contents = None


UnresolvedPage = typ.Union[int, PDFObjRef]
"""A reference to a page that is *either* a page number, or a PDF object ID."""


class Outline(ObjectWithPos):
    """
    A PDF outline (also known as a bookmark).

    Outlines are used to navigate the PDF, and are often headings in the
    document's table of contents. A single outline has a title (name), and a
    target location in the PDF (page and X/Y coordinates). Initially the page is
    referred to by reference, but the reference is unresolved -- it is either a
    page number, or a PDF object ID. While rendering the PDF, the page is
    resolved to a Page object, and the pos attribute is updated.
    """

    def __init__(
        self,
        title: str,
        pageref: UnresolvedPage,
        target: typ.Optional[typ.Tuple[float, float]]
    ):
        super().__init__()
        self.title = title
        self.pageref = pageref
        self.target = target

    def __repr__(self) -> str:
        return '<Outline \'%s\' %r>' % (self.title, self.pos)

    def resolve(self, page: Page) -> None:
        """Resolve our page reference to the given page, and update our position."""
        assert self.pos is None
        if isinstance(self.pageref, PDFObjRef):
            assert self.pageref.objid == page.objid
        else:
            assert self.pageref == page.pageno

        if self.target is None:
            # XXX: "first" point on the page, assuming left-to-right top-to-bottom order
            (targetx, targety) = (page.mediabox.x0, page.mediabox.y1)
        else:
            (targetx, targety) = self.target

        self.pos = Pos(page, targetx, targety)


class Document:
    """
    A fully-extracted PDF document.

    This is really just a list of pages and some helpers.

    Attributes:
        pages   An ordered list of Page objects, indexed by zero-based page number.
    """

    pages: typ.List[Page]

    def __init__(self) -> None:
        self.pages = []

    def iter_annots(self, *, include_replies: bool = False) -> typ.Iterator[Annotation]:
        """
        Iterate over all the annotations in the document.

        Only the primary annotation for a group is included.
        Replies are included only if include_replies is True.
        """

        for p in self.pages:
            for a in p.annots:
                if not a.is_group_child and (include_replies or not a.in_reply_to):
                    yield a

    def nearest_outline(
        self,
        pos: Pos
    ) -> typ.Optional[Outline]:
        """Return the first outline occuring prior to the given position, in reading order."""

        # Search pages backwards from the given pos
        for pageno in range(pos.page.pageno, -1, -1):
            page = self.pages[pageno]
            assert page.pageno == pageno

            # Outlines are pre-sorted, so we can use bisect to find the first outline < pos
            idx = bisect.bisect(page.outlines, ObjectWithPos(pos))
            if idx:
                return page.outlines[idx - 1]

        return None


class RGB(typ.NamedTuple):
    red: float
    green: float
    blue: float

    def ashex(self) -> str:
        "Return a 6-character string representing the 24-bit hex code for this colour."
        red_hex = format(int(self.red * 255), '02x')
        green_hex = format(int(self.green * 255), '02x')
        blue_hex = format(int(self.blue * 255), '02x')
        return red_hex + green_hex + blue_hex

    def __str__(self) -> str:
        return f"RGB({self.ashex()})"



#!/usr/bin/env python3

# This script, which is not part of the pdfannots package, allows pdfannots
# to by run directly from a source tree clone.

if __name__ == '__main__':
    import pdfannots.cli
    pdfannots.cli.main()
